__all__ = [
    'find_tci_path',
    'find_img_folder_in_product',
    'find_jp2_files_in_folder',
    'find_detections_png_files',
    'find_tiles_files',
    'find_matching_jp2_files',
    'remove_duplicate_files'
]
from .find_folders import (find_tci_path,
                           find_img_folder_in_product,
                           find_jp2_files_in_folder,
                           find_detections_png_files,
                           find_tiles_files,
                           find_matching_jp2_files,
                           copy_file_to_folder,
                           remove_duplicate_files)


import requests
from datetime import datetime
import pandas as pd
from ..polygon import coords_to_str

def products_by_polygon(polygon_coords:list, data_start:str, data_end:str,collection_name: str = "SENTINEL-2", top = 1000):
    polygon_coords = coords_to_str(polygon_coords)
    url = f"https://catalogue.dataspace.copernicus.eu/odata/v1/Products?$filter=Collection/Name eq '{collection_name}' and OData.CSC.Intersects(area=geography'SRID=4326;POLYGON({polygon_coords})') and ContentDate/Start gt {data_start} and ContentDate/Start lt {data_end}&$top={top}"
    json = requests.get(
        url).json()
    df = pd.DataFrame.from_dict(json['value'])
    pd.set_option('display.max_colwidth', None)
    print(f'Число результатов:{len(df)}')
    df_dict = df.to_dict()
    return df_dict

def products_by_coords(lat, lon, data_start, data_end, collection_name: str = "SENTINEL-2", top: int = 1000):
    url = f"https://catalogue.dataspace.copernicus.eu/odata/v1/Products?$filter=Collection/Name eq '{collection_name}' and OData.CSC.Intersects(area=geography'SRID=4326;POINT({lon} {lat})') and ContentDate/Start gt {data_start} and ContentDate/Start lt {data_end}&$top={top}"
    json = requests.get(
        url).json()
    df = pd.DataFrame.from_dict(json['value'])
    pd.set_option('display.max_colwidth', None)
    print(f'Число результатов:{len(df)}')
    df_dict = df.to_dict()
    return df_dict



from datetime import datetime

def filter_products_by_l2a(product_dict):
    filtered_products_id = {key: value for key, value in product_dict['Name'].items() if "L2A" in value}
    filtered_indecies = [key for key in filtered_products_id.keys()]

    filtered_products = get_data_by_index(product_dict, filtered_indecies)
    print(f"Число результатов l2a:{len(filtered_products['Name'])}" )
    return filtered_products


def filter_latest_product(product_dict):
    product_dates = product_dict['ContentDate']
    parsed_dates = {
        index: datetime.fromisoformat(value['Start'].rstrip('Z'))
        for index, value in product_dates.items()
    }
    latest_index = [max(parsed_dates, key=parsed_dates.get)]

    latest_product = get_data_by_index(product_dict, latest_index)
    flat_data = get_flat_data(latest_product)
    return flat_data

# Вспомогательная функция для получения всех записей с нужным индексом
def get_data_by_index(data_dict, filtered_indexies):
    filtered_data = {}
    for key_name, value in data_dict.items():
        # print(f'{key_name} : {value}')
        filtered_data_items = {key: value for key, value in data_dict[key_name].items() if key in filtered_indexies}
        filtered_data[key_name] = filtered_data_items
    return filtered_data

def get_flat_data(data):
    # Определяем индекс (предполагаем, что все ключи имеют одинаковый индекс)
    index = next(iter(data.values())).keys().__iter__().__next__()

    # Преобразуем словарь, убирая индекс
    flat_data = {key: value[index] for key, value in data.items()}
    return flat_data


class NotUpdateToken(Exception):
    def __init__(self, message = 'Не удалось обновить токен'):
        self.message = message
        super().__init__(self.message)

class ProductNotFoundError(Exception):
    def __init__(self, message = 'Не удалось найти файл'):
        self.message = message
        super().__init__(self.message)

class ProductNotLoadedError(Exception):
    def __init__(self, message = 'Не удалось скачать файл'):
        self.message = message
        super().__init__(self.message)

class TooManyFoldersError(Exception):
    def __init__(self, message = "Найдено слишком много папок"):
        self.message = message
        super().__init__(self.message)


__all__ = [
    'crop_image_by_polygon',
    'remove_black_borders',
    'split_jp2_into_tiles',
    'resample',
    'convert_files_to_png',
    'convert_jp2_to_png'
]

from .cropper import crop_image_by_polygon, remove_black_borders, split_jp2_into_tiles
from .format import resample
from .jp2_png_converter import convert_files_to_png, convert_jp2_to_png


from senphora.loader import load_product
from senphora.utils import unzip
from senphora.finder import find_img_folder_in_product, find_tci_path, find_jp2_files_in_folder
from senphora.image import crop_image_by_polygon, remove_black_borders, resample, split_jp2_into_tiles, convert_files_to_png
from senphora.logger import logger
from dataclasses import dataclass
import os


def load_and_unzip(access_token, product):
    """
    Load and unzip product
    :param access_token:
    :param product:
    :return:
    """
    zip_file_path, extract_dir = load_product(access_token, product)
    unzip(zip_file_path, extract_dir)
    return extract_dir

def crop_folder(product_folder):
    """
    Prepare folder for cropped image
    :param product_folder: Folder with product.zip and unpacked folder
    :return:
    """
    output_image_folder = os.path.join(product_folder, "cropped")
    cropped_name = 'cropped.jp2'
    blacked_name = 'blacked.jp2'
    cropped_image_path = os.path.join(output_image_folder, cropped_name)
    blacked_image_path = os.path.join(output_image_folder, blacked_name)
    os.makedirs(output_image_folder, exist_ok=True)
    logger.info(f"Папка для обрезки подготовлена: {output_image_folder}")
    return cropped_image_path, blacked_image_path

def tiles_folder(product_folder):
    """
    Prepare tiles folder for cropped image
    :param product_folder: Folder with product.zip and unpacked folder
    :return:
    """
    output_dir = os.path.join(product_folder, 'tiles')
    base_tiles_name = os.path.basename(os.path.normpath(product_folder)) # - название предпоследней папки
    os.makedirs(output_dir, exist_ok=True)
    logger.info(f"Папка для тайлов подготовлена: {output_dir}")
    return output_dir, base_tiles_name

def png_folder(product_folder):
    """
    Prepare folder for png_tiles
    :param product_folder: Folder with product.zip and unpacked folder
    :return:
    """
    output_dir = os.path.join(product_folder, 'png_data')
    os.makedirs(output_dir, exist_ok=True)
    logger.info(f"Папка для png подготовлена: {output_dir}")
    return output_dir

def tci_image_path(product_folder):
    """
    Prepare path to tci image
    :param product_folder: Folder with product.zip and unpacked folder
    :return:
    """
    tci_folder = find_img_folder_in_product(product_folder)
    tci_image = find_tci_path(tci_folder)
    logger.info(f"TCI изображение найдено: {tci_image}")
    return tci_image

@dataclass
class PrepareFolders:
    """
    class for preparing folders
    :param cropped_img: path to cropped image
    :param blacked_img: path to blacked image
    :param tci_img: path to tci image
    :param tiles_dir: path to tiles folder
    :param tiles_base_name: base name of tiles folder
    :param png_dir: path to png folder
    """
    cropped_img: str
    blacked_img: str
    tci_img: str
    tiles_base_name: str
    tiles_dir: str
    png_dir: str


def prepare_folders_neuro(product_folder: str) -> PrepareFolders:
    """
    Prepare folders for neural network
    :param product_folder:
    :return:
    """
    cropped_img_path, blacked_img_path = crop_folder(product_folder)
    tiles_dir, base_tile_name = tiles_folder(product_folder)
    png_dir = png_folder(product_folder)
    tci_image = tci_image_path(product_folder)
    prepared_folders = PrepareFolders(
        cropped_img = cropped_img_path,
        blacked_img = blacked_img_path,
        tiles_dir = tiles_dir,
        tiles_base_name = base_tile_name,
        png_dir = png_dir,
        tci_img = tci_image
    )
    return prepared_folders

def prepare_img_neuro(prepared_folders: PrepareFolders, polygon_coords: list):
    """
    Prepare image for neural network
    :param prepared_folders: PreparedFolders class
    :param polygon_coords: list of coordinates of polygon
    :return:
    """
    crop_image_by_polygon(prepared_folders.tci_img, prepared_folders.cropped_img, polygon_coords)
    remove_black_borders(prepared_folders.cropped_img, prepared_folders.blacked_img)
    resample(prepared_folders.blacked_img)
    split_jp2_into_tiles(prepared_folders.blacked_img, prepared_folders.tiles_dir, prepared_folders.tiles_base_name)
    jp2_files = find_jp2_files_in_folder(prepared_folders.tiles_dir)
    png_dir = convert_files_to_png(jp2_files, prepared_folders.png_dir)
    return png_dir


import zipfile
import os
from tqdm import tqdm
from senphora.logger import logger

def unzip(zip_file_path, extract_dir):
    """
    Распаковывает архив в указанную директорию.

    :param zip_file_path: Путь к ZIP-архиву.
    :param extract_dir: Директория, куда будут извлечены файлы.
    """
    # Проверяем, существует ли архив
    if not os.path.exists(zip_file_path):
        raise FileNotFoundError(f"Архив {zip_file_path} не найден.")

    # Определяем директорию для распаковки
    unpack_dir = os.path.join(os.getcwd(), extract_dir, 'product_files')

    # Создаем директорию для извлечения, если она не существует
    os.makedirs(unpack_dir, exist_ok=True)

    # Распаковываем архив с прогресс-баром
    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
        # Получаем список всех файлов в архиве
        total_files = len(zip_ref.infolist())

        with tqdm(iterable=zip_ref.infolist(), total=total_files, unit="file", desc="Распаковка") as pbar:
            for file in pbar:
                zip_ref.extract(file, path=unpack_dir)
                pbar.set_description(f"Распаковка {file.filename}")
    logger.info(f"Архив {zip_file_path} успешно распакован в {unpack_dir}.")
    return unpack_dir



import os

import numpy as np
import rasterio
from tqdm import tqdm
from ..polygon import read_polygon
from pyproj import Transformer
from rasterio.crs import CRS
from rasterio.mask import mask
from rasterio.windows import Window
from shapely.geometry import Polygon, mapping
from .format import resample
from senphora.logger import logger


def crop_image_by_polygon(image_path, output_image_path, polygon_coords):
    """
    Crop image by polygon coordinates
    :param image_path: путь до изображения
    :param product_folder: корневая папка продукта
    :param polygon_coords: координаты полигона
    :return:
    """
    # Преобразование координат полигона из EPSG:4326 в EPSG:32636
    with rasterio.open(image_path) as src:
        raster_crs = src.crs  # Получаем CRS растра (должно быть EPSG:32636)
        # Преобразование координат
        transformer = Transformer.from_crs(CRS.from_epsg(4326), raster_crs, always_xy=True)
        transformed_coords = [transformer.transform(x, y) for x, y in polygon_coords]
        polygon_geom = Polygon(transformed_coords)

        # Проверка пересечения полигонов
        raster_bounds = src.bounds
        if not Polygon.from_bounds(*raster_bounds).intersects(polygon_geom):
            raise ValueError("Полигон не пересекается с растровым изображением.")

        # Обрезка изображения по полигону
        geoms = [mapping(polygon_geom)]
        out_image, out_transform = mask(src, geoms, crop=True)

        # Копирование метаданных
        out_meta = src.meta.copy()
        out_meta.update({
            "driver": "GTiff",  # Формат выходного файла
            "height": out_image.shape[1],  # Высота обрезанного изображения
            "width": out_image.shape[2],  # Ширина обрезанного изображения
            "transform": out_transform,  # Новая матрица трансформации
            "count": src.count,  # Число каналов (должно быть 3 для RGB)
            "dtype": src.dtypes[0]  # Тип данных (например, uint8)
        })

    # Сохранение обрезанного изображения
    with rasterio.open(output_image_path, "w", **out_meta) as dest:
        dest.write(out_image)  # Записываем все каналы

    logger.info(f"Обрезанное изображение сохранено: {output_image_path}")
    return output_image_path

def crop_image_by_lat_lon(image_path, output_image_path, lat, lon, img_size=640):
    with rasterio.open(image_path) as src:
        raster_crs = src.crs
        print(f"Растровая система координат: {raster_crs}")
        # Преобразование координат из EPSG:4326 в систему координат растра
        transformer = Transformer.from_crs(CRS.from_epsg(4326), raster_crs, always_xy=True)
        x_center, y_center = transformer.transform(lon, lat)
        # Проверка, что центр попадает внутрь растра
        if not (
                src.bounds.left <= x_center <= src.bounds.right and src.bounds.bottom <= y_center <= src.bounds.top):
            raise ValueError("Указанный центр вне области растрового изображения.")
        # Размер стороны квадрата в пикселях
        square_size = img_size
        half_size = square_size // 2

        # Определяем индексы пикселей для обрезки
        row_start = int(src.height - (y_center - src.bounds.bottom) / src.res[1] - half_size)
        row_stop = row_start + square_size
        col_start = int((x_center - src.bounds.left) / src.res[0] - half_size)
        col_stop = col_start + square_size

        # Корректируем границы, чтобы они не выходили за пределы растра
        row_start_clipped = max(0, row_start)
        row_stop_clipped = min(src.height, row_stop)
        col_start_clipped = max(0, col_start)
        col_stop_clipped = min(src.width, col_stop)

        # Создаем окно обрезки
        window = Window(col_start_clipped, row_start_clipped,
                        col_stop_clipped - col_start_clipped,
                        row_stop_clipped - row_start_clipped)

        # Читаем данные из растра
        out_image = src.read(window=window)

        # Создаем новое изображение размером 640x640, заполненное черным цветом
        full_image = np.zeros((src.count, square_size, square_size), dtype=src.dtypes[0])

        # Вычисляем смещение для вставки обрезанного фрагмента
        row_offset = max(0, -row_start)
        col_offset = max(0, -col_start)

        # Вставляем обрезанный фрагмент в новое изображение
        full_image[:, row_offset:row_offset + out_image.shape[1],
        col_offset:col_offset + out_image.shape[2]] = out_image

        # Вычисляем новую матрицу трансформации
        new_transform = rasterio.transform.from_bounds(
            src.bounds.left + col_start * src.res[0],
            src.bounds.bottom + (src.height - row_stop) * src.res[1],
            src.bounds.left + col_stop * src.res[0],
            src.bounds.bottom + (src.height - row_start) * src.res[1],
            width=square_size,
            height=square_size
        )

        # Копируем метаданные
        out_meta = src.meta.copy()
        out_meta.update({
            "driver": "GTiff",  # Формат выходного файла
            "height": square_size,  # Высота обрезанного изображения
            "width": square_size,  # Ширина обрезанного изображения
            "transform": new_transform,  # Новая матрица трансформации
            "count": src.count,  # Число каналов (должно быть 3 для RGB)
            "dtype": src.dtypes[0]  # Тип данных (например, uint8)
        })

        # Сохраняем обрезанное изображение
        with rasterio.open(output_image_path, "w", **out_meta) as dest:
            dest.write(full_image)

        print(f"Тайл сохранен: {output_image_path}")
        return output_image_path

def remove_black_borders(input_image_path, output_img_path, nodata_value=0):
    with rasterio.open(input_image_path) as src:
        # Чтение данных растра
        image = src.read()
        height, width = src.height, src.width
        transform = src.transform

        # Поиск непустых пикселей
        mask = (image != nodata_value).any(axis=0)  # Маска непустых пикселей
        rows = mask.any(axis=1)  # Непустые строки
        cols = mask.any(axis=0)  # Непустые столбцы

        # Определение границ непустой области
        row_start, row_end = rows.argmax(), len(rows) - rows[::-1].argmax()
        col_start, col_end = cols.argmax(), len(cols) - cols[::-1].argmax()

        # Создание окна обрезки
        window = Window(col_start, row_start, col_end - col_start, row_end - row_start)

        # Обрезка данных и обновление трансформации
        cropped_image = src.read(window=window)
        updated_transform = rasterio.windows.transform(window, transform)

        # Обновление метаданных
        out_meta = src.meta.copy()
        out_meta.update({
            "height": cropped_image.shape[1],
            "width": cropped_image.shape[2],
            "transform": updated_transform,
            "nodata": nodata_value
        })

        # Сохранение обрезанного изображения
        with rasterio.open(output_img_path, "w", **out_meta) as dest:
            dest.write(cropped_image)
    logger.info(f"Черные пиксели преобразованы. Результат сохранен в {output_img_path}")
    return output_img_path


def split_jp2_into_tiles(input_img_path, tiles_dir, base_tile_name, tile_size=640):
    # Открываем JP2 файл
    with rasterio.open(input_img_path) as src:
        width, height = src.width, src.height
        transform = src.transform
        crs = src.crs

        # Создаем трансформер для преобразования в WGS84
        transformer = Transformer.from_crs(crs, "EPSG:4326", always_xy=True)

        # Вычисляем новые размеры для кратности tile_size
        new_width = ((width + tile_size - 1) // tile_size) * tile_size
        new_height = ((height + tile_size - 1) // tile_size) * tile_size

        # Если размеры уже кратны tile_size, ничего не меняем
        if new_width == width and new_height == height:
            padded_data = src.read()
            updated_transform = transform
        else:
            # Читаем исходные данные
            data = src.read()

            # Создаем массив для расширенного изображения
            padded_data = np.zeros((data.shape[0], new_height, new_width), dtype=data.dtype)

            # Копируем исходные данные в новый массив
            padded_data[:, :height, :width] = data

            # Обновляем матрицу трансформации
            updated_transform = rasterio.Affine(
                transform.a, transform.b, transform.c,
                transform.d, transform.e, transform.f
            )

        # Вычисляем общее количество блоков
        total_blocks = (height // tile_size + (height % tile_size > 0)) * \
                       (width // tile_size + (width % tile_size > 0))
        # Проходим по всем блокам с помощью tqdm для отображения прогресса
        block_count = 0
        # Проходим по всем блокам с помощью tqdm для отображения прогресса
        with tqdm(total=total_blocks, desc="Обработка блоков", unit="блок") as pbar:
            for row in range(0, height, tile_size):
                for col in range(0, width, tile_size):
                    block_count += 1
                    # Определяем размер текущего блока
                    window = Window(col, row, min(tile_size, new_width - col), min(tile_size, new_height - row))

                    # Читаем данные блока
                    tile_data = padded_data[:, window.row_off:window.row_off + window.height,
                                window.col_off:window.col_off + window.width]

                    # Получаем географические координаты верхнего левого угла блока
                    ul_x, ul_y = rasterio.transform.xy(updated_transform, row, col)

                    # Преобразуем координаты в WGS84
                    wgs84_x, wgs84_y = transformer.transform(ul_x, ul_y)

                    # Формируем имя файла с координатами
                    tile_name = f"tile_{base_tile_name}_{wgs84_x:.6f}_{wgs84_y:.6f}.jp2"
                    output_tile = os.path.join(tiles_dir, tile_name)

                    # Копируем метаданные
                    out_meta = src.meta.copy()
                    out_meta.update({
                        "driver": "GTiff",
                        "height": window.height,
                        "width": window.width,
                        "transform": rasterio.windows.transform(window, updated_transform),
                    })

                    # Сохраняем блок
                    with rasterio.open(output_tile, "w", **out_meta) as dst:
                        dst.write(tile_data)

                    # Обновляем прогресс-бар
                    pbar.update(1)
                # print(f"Сохранен блок: {output_path}")
    logger.info(f"Разбиение на тайлы завершено. Тайлы сохранены в папку: {tiles_dir}")

if __name__ == "__main__":
    product_folder = '../data/2025/01/10/S2B_T43PDN_MSIL2A_05_21_09'
    filename = '../polygon_files/india_abma.geojson'
    polygon_coords = read_polygon(filename)
    crop_image_by_polygon(product_folder, polygon_coords)
    final_cropped_image = remove_black_borders(product_folder)
    resample(final_cropped_image)
    split_jp2_into_tiles(product_folder)
